from zrb import CmdTask, StrInput, Env, EnvFile, runner, python_task
from zrb.builtin.group import project_group
from ._common import CURRENT_DIR, RESOURCE_DIR, APP_TEMPLATE_ENV_FILE_NAME
from .frontend import build_snake_zrb_app_name_frontend_once
from .local import prepare_snake_zrb_app_name_backend
import os

###############################################################################
# Env file Definitions
###############################################################################

app_env_file = EnvFile(
    env_file=APP_TEMPLATE_ENV_FILE_NAME, prefix='TEST_ZRB_ENV_PREFIX'
)

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    icon='🧪',
    name='remove-kebab-zrb-app-name-test-db',
    group=project_group,
    runner=runner
)
def remove_snake_zrb_app_name_test_db(*args, **kwargs):
    test_db_file_path = os.path.join(RESOURCE_DIR, 'test.db')
    if os.path.isfile(test_db_file_path):
        os.remove(test_db_file_path)


test_snake_zrb_app_name = CmdTask(
    icon='🚤',
    name='test-kebab-zrb-app-name',
    group=project_group,
    inputs=[
        StrInput(
            name='kebab-zrb-app-name-test',
            description='Specific test case (i.e., test/file.py::test_name)',
            prompt='Test (i.e., test/file.py::test_name)',
            default=''
        )
    ],
    upstreams=[
        build_snake_zrb_app_name_frontend_once,
        prepare_snake_zrb_app_name_backend,
        remove_snake_zrb_app_name_test_db,
    ],
    cwd=RESOURCE_DIR,
    env_files=[app_env_file],
    envs=[
        Env(
            name='APP_BROKER_TYPE',
            os_name='TEST_APP_BROKER_TYPE',
            default='mock'
        ),
        Env(
            name='APP_DB_CONNECTION',
            os_name='TEST_APP_DB_CONNECTION',
            default='sqlite:///test.db'
        ),
        Env(
            name='APP_AUTH_ADMIN_ACTIVE',
            os_name='',
            default='true'
        ),
        Env(
            name='APP',
            os_name='APP_ENABLE_OTEL',
            default='false'
        ),
    ],
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'app-test.sh'),
    retry=0
)
runner.register(test_snake_zrb_app_name)



from typing import TypeVar, Type
from pydantic import BaseModel
from core.repo.repo import Repo
from core.messagebus.messagebus import Publisher
from core.model.repo_model import RepoModel
from module.log.schema.activity import ActivityData
import jsons


Schema = TypeVar('Schema', bound=BaseModel)
SchemaData = TypeVar('SchemaData', bound=BaseModel)
SchemaResult = TypeVar('SchemaResult', bound=BaseModel)


class HistoricalRepoModel(RepoModel[Schema, SchemaData, SchemaResult]):
    schema_result_cls: Type[SchemaResult]
    log_entity_name: str

    def __init__(self, repo: Repo[Schema, SchemaData], publisher: Publisher):
        super().__init__(repo)
        self.publisher = publisher

    async def insert(self, data: SchemaData) -> Schema:
        result = await super().insert(data)
        await self._publish_activity(action='insert', result=result)
        return result

    async def update(self, id: str, data: SchemaData) -> Schema:
        result = await self.repo.update(id, data)
        await self._publish_activity(action='update', result=result)
        return result

    async def delete(self, id: str) -> Schema:
        result = await self.repo.delete(id)
        await self._publish_activity(action='delete', result=result)
        return result

    async def _publish_activity(self, action: str, result: Schema):
        activity_data = ActivityData(
            action=action,
            entity=self.log_entity_name,
            data=jsons.dumps(result.dict())
        )
        await self.publisher.publish('log_new_activity', activity_data.dict())



from component.messagebus import publisher
from module.snake_zrb_module_name.component.repo.snake_zrb_entity_name_repo import (
    snake_zrb_entity_name_repo
)
from module.snake_zrb_module_name.entity.snake_zrb_entity_name.model import (
    PascalZrbEntityNameModel
)

snake_zrb_entity_name_model: PascalZrbEntityNameModel = PascalZrbEntityNameModel(
    snake_zrb_entity_name_repo, publisher
)



from typing import Any, Mapping
from typeguard import typechecked
from ..common.task_input import project_dir_input, project_name_input
from ..project_task.task_factory import create_ensure_project_tasks
from ...group import project_group
from ....task.cmd_task import CmdTask
from ....task.decorator import python_task
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from ....config.config import version

import os

CURRENT_DIR = os.path.dirname(__file__)
SYSTEM_USER = os.getenv('USER', 'incognito')
IMAGE_DEFAULT_NAMESPACE = os.getenv(
    'PROJECT_IMAGE_DEFAULT_NAMESPACE', 'docker.io/' + SYSTEM_USER
)

###############################################################################
# Replacement Mutator Definitions
###############################################################################


@typechecked
def copy_resource_replacement_mutator(
    task: ResourceMaker, replacements: Mapping[str, str]
) -> Mapping[str, str]:
    replacements['zrbBaseProjectDir'] = os.path.basename(
        replacements.get('zrbProjectDir', '')
    )
    if replacements.get('zrbProjectName', '') == '':
        replacements['zrbProjectName'] = replacements.get(
            'zrbBaseProjectDir', ''
        )
    return replacements


###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=[project_dir_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    if os.path.isfile(os.path.join(project_dir, 'zrb_init.py')):
        raise Exception(f'Project directory already exists: {project_dir}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        project_name_input
    ],
    upstreams=[validate],
    replacements={
        'zrbProjectDir': '{{input.project_dir}}',
        'zrbProjectName': '{{input.project_name}}',
        'zrbImageDefaultNamespace': IMAGE_DEFAULT_NAMESPACE,
        'zrbVersion': version,
    },
    replacement_mutator=copy_resource_replacement_mutator,
    template_path=os.path.join(CURRENT_DIR, 'template'),
    destination_path='{{input.project_dir}}',
    excludes=[
        '*/__pycache__',
    ]
)

ensure_project_tasks = create_ensure_project_tasks(
    upstreams=[copy_resource]
)

create_project = CmdTask(
    name='create',
    group=project_group,
    upstreams=[ensure_project_tasks],
    inputs=[project_dir_input],
    cmd=[
        'set -e',
        'cd "{{input.project_dir}}"',
        'if [ ! -d .git ]',
        'then',
        '  echo Initialize project git repository',
        '  git init',
        'fi',
        'echo "Happy coding :)"',
    ]
)
runner.register(create_project)



from typing import Any, List
from zrb import CmdTask, DockerComposeTask, Task, Env, EnvFile, runner
from zrb.builtin.group import project_group
from ._common import (
    CURRENT_DIR, APP_DIR, APP_TEMPLATE_ENV_FILE_NAME, RESOURCE_DIR,
    skip_local_microservices_execution, rabbitmq_checker,
    rabbitmq_management_checker, redpanda_console_checker,
    kafka_outside_checker, kafka_plaintext_checker, pandaproxy_outside_checker,
    pandaproxy_plaintext_checker, app_local_checker, local_input, https_input,
    run_mode_input, enable_monitoring_input, host_input, app_enable_otel_env
)
from .image import image_input
from .frontend import build_snake_zrb_app_name_frontend
from .container import remove_snake_zrb_app_name_container, compose_env_file
from .local_microservices import get_start_microservices
import os

###############################################################################
# Functions
###############################################################################


def setup_support_compose_profile(*args: Any, **kwargs: Any) -> str:
    task: Task = kwargs.get('_task')
    env_map = task.get_env_map()
    compose_profiles: List[str] = []
    broker_type = env_map.get('APP_BROKER_TYPE', 'rabbitmq')
    if broker_type in ['rabbitmq', 'kafka']:
        compose_profiles.append(broker_type)
    if kwargs.get('enable_snake_zrb_app_name_monitoring', False):
        compose_profiles.append('monitoring')
    compose_profile_str = ','.join(compose_profiles)
    return f'export COMPOSE_PROFILES={compose_profile_str}'


def skip_support_container_execution(*args: Any, **kwargs: Any) -> bool:
    return not kwargs.get('local_snake_zrb_app_name', True)


def skip_local_monolith_execution(*args: Any, **kwargs: Any) -> bool:
    if not kwargs.get('local_snake_zrb_app_name', True):
        return True
    return kwargs.get('snake_zrb_app_name_run_mode', 'monolith') != 'monolith'


###############################################################################
# Env file Definitions
###############################################################################

app_env_file = EnvFile(
    env_file=APP_TEMPLATE_ENV_FILE_NAME, prefix='ZRB_ENV_PREFIX'
)

###############################################################################
# Task Definitions
###############################################################################

init_snake_zrb_app_name_support_container = DockerComposeTask(
    icon='🔥',
    name='init-kebab-zrb-app-name-support-container',
    inputs=[
        local_input,
        host_input,
        image_input,
    ],
    skip_execution=skip_support_container_execution,
    upstreams=[
        remove_snake_zrb_app_name_container
    ],
    cwd=RESOURCE_DIR,
    setup_cmd=setup_support_compose_profile,
    compose_cmd='up',
    compose_flags=['-d'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    env_files=[compose_env_file]
)

start_snake_zrb_app_name_support_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-zrb-app-name-support-container',
    description='Start human readable zrb app name container',
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
        image_input,
    ],
    skip_execution=skip_support_container_execution,
    upstreams=[init_snake_zrb_app_name_support_container],
    cwd=RESOURCE_DIR,
    setup_cmd=setup_support_compose_profile,
    compose_cmd='logs',
    compose_flags=['-f'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    env_files=[compose_env_file],
    checkers=[
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)

prepare_snake_zrb_app_name_backend = CmdTask(
    icon='🚤',
    name='prepare-kebab-zrb-app-name-backend',
    description='Prepare backend for human readable zrb app name',
    group=project_group,
    cwd=APP_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-prepare-backend.sh'),
    ]
)
runner.register(prepare_snake_zrb_app_name_backend)

start_monolith_snake_zrb_app_name = CmdTask(
    icon='🚤',
    name='start-monolith-kebab-zrb-app-name',
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        https_input
    ],
    skip_execution=skip_local_monolith_execution,
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[app_env_file],
    envs=[app_enable_otel_env],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-start.sh'),
    ],
    checkers=[
        app_local_checker,
    ]
)

start_snake_zrb_app_name_gateway = CmdTask(
    icon='🚪',
    name='start-kebab-zrb-app-name-gateway',
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        https_input
    ],
    skip_execution=skip_local_microservices_execution,
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[
        app_env_file,
    ],
    envs=[
        Env(name='APP_DB_AUTO_MIGRATE', default='false', os_name=''),
        Env(name='APP_ENABLE_EVENT_HANDLER', default='false', os_name=''),
        Env(name='APP_ENABLE_RPC_SERVER', default='false', os_name=''),
        app_enable_otel_env,
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-start.sh'),
    ],
    checkers=[
        app_local_checker,
    ]
)

start_microservices = get_start_microservices(
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ]
)

start_snake_zrb_app_name = Task(
    icon='🚤',
    name='start-kebab-zrb-app-name',
    description='Start human readable zrb app name',
    group=project_group,
    upstreams=[
        start_monolith_snake_zrb_app_name,
        start_snake_zrb_app_name_gateway,
    ] + start_microservices,
    run=lambda *args, **kwargs: kwargs.get('_task').print_out('🆗')
)
runner.register(start_snake_zrb_app_name)



from zrb import CmdTask, Env, EnvFile, HTTPChecker, runner
from zrb.builtin.group import project_group
from ._common import (
    CURRENT_DIR, APP_DIR, TEMPLATE_ENV_FILE_NAME,
    local_input, host_input, https_input
)
import os

###############################################################################
# Env File Definitions
###############################################################################

app_env_file = EnvFile(env_file=TEMPLATE_ENV_FILE_NAME, prefix='ZRB_ENV_PREFIX')

###############################################################################
# Task Definitions
###############################################################################

start_snake_zrb_app_name = CmdTask(
    icon='🚤',
    name='start-kebab-zrb-app-name',
    description='Start human readable zrb app name',
    group=project_group,
    inputs=[
        local_input,
        host_input,
        https_input
    ],
    skip_execution='{{not input.local_snake_zrb_app_name}}',
    cwd=APP_DIR,
    env_files=[app_env_file],
    envs=[
        Env(
            name='APP_PORT',
            os_name='ZRB_ENV_PREFIX_APP_PORT',
            default='zrbAppHttpPort'
        )
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'app-activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-start.sh'),
    ],
    checkers=[
        HTTPChecker(
            name='check-kebab-zrb-app-name',
            host='{{input.snake_zrb_app_name_host}}',
            port='{{env.APP_PORT}}',
            is_https='{{input.snake_zrb_app_name_https}}'
        )
    ]
)
runner.register(start_snake_zrb_app_name)



from typing import List
from core.schema import BaseDateTimeSchema, BaseCountSchema
from module.auth.schema.permission import Permission


class GroupBase(BaseDateTimeSchema):
    name: str
    description: str


class GroupData(GroupBase):
    permissions: List[str]


class Group(GroupBase):
    id: str
    permissions: List[Permission] = []

    class Config:
        orm_mode = True


class GroupResult(BaseCountSchema):
    data: List[Group]



from ..config.config import logging_level
from .accessories.color import colored
import logging

# create logger
logger = logging.getLogger('zrb')
logger.setLevel(logging_level)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging_level)

# create formatter
formatter = logging.Formatter(
    colored('%(levelname)-8s %(asctime)s', attrs=['dark']) + ' %(message)s'
)

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)



from abc import ABC, abstractmethod
from core.error import HTTPAPIException
from module.auth.schema.token import RefreshTokenData
from datetime import datetime, timedelta
from jose import jwt

import jsons


class RefreshTokenUtil(ABC):

    @abstractmethod
    def encode(self, token_data: RefreshTokenData) -> str:
        pass

    @abstractmethod
    def decode(self, token_str: str) -> RefreshTokenData:
        pass


class JWTRefreshTokenUtil(RefreshTokenUtil):

    def __init__(self, secret_key: str, algorithm: str = 'HS256'):
        self.secret_key = secret_key
        self.algorithm = algorithm

    def encode(self, data: RefreshTokenData) -> str:
        expire_time = datetime.utcnow() + timedelta(
            seconds=data.expire_seconds
        )
        sub = jsons.dumps(data.dict())
        data_dict = {'sub': sub, 'exp': expire_time}
        encoded_jwt = jwt.encode(
            data_dict, self.secret_key, algorithm=self.algorithm
        )
        return encoded_jwt

    def decode(self, token: str) -> RefreshTokenData:
        try:
            decoded_data = jwt.decode(
                token, self.secret_key, algorithms=[self.algorithm]
            )
            sub = jsons.loads(decoded_data['sub'])
            token_data = RefreshTokenData.parse_obj(sub)
            expire_time = decoded_data['exp']
            token_data.expire_seconds = ((
                datetime.fromtimestamp(expire_time) - datetime.utcnow()
            ).total_seconds())
            if token_data.expire_seconds < 0:
                raise HTTPAPIException(422, 'Expired token')
            return token_data
        except jwt.JWTError:
            raise HTTPAPIException(422, 'Invalid token')



from typeguard import typechecked
from ....task.task import Task
from ....helper.codemod.add_import_module import add_import_module
from ....helper.codemod.append_code_to_function import append_code_to_function
from ....helper.file.text import read_text_file_async, write_text_file_async
import os


@typechecked
async def register_api(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str
):
    module_api_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'api.py'
    )
    register_function_path = '.'.join([
        'module', snake_module_name, 'entity', snake_entity_name, 'api'
    ])
    register_function = f'register_{snake_entity_name}_api'
    task.print_out(f'Read code from: {module_api_file_path}')
    code = await read_text_file_async(module_api_file_path)
    task.print_out(
        f'Add import "register_api" as "{register_function}" ' +
        f'from "{register_function_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=register_function_path,
        resource='register_api',
        alias=register_function
    )
    task.print_out(f'Add "{register_function}" call to the code')
    code = append_code_to_function(
        code=code,
        function_name='register_api',
        new_code=f'{register_function}(logger, app, authorizer, rpc_caller, publisher)' # noqa
    )
    task.print_out(f'Write modified code to: {module_api_file_path}')
    await write_text_file_async(module_api_file_path, code)


@typechecked
async def register_rpc(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str
):
    module_rpc_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'rpc.py'
    )
    register_function_path = '.'.join([
        'module', snake_module_name, 'entity', snake_entity_name, 'rpc'
    ])
    register_function = f'register_{snake_entity_name}_rpc'
    task.print_out(f'Read code from: {module_rpc_file_path}')
    code = await read_text_file_async(module_rpc_file_path)
    task.print_out(
        f'Add import "register_rpc" as "{register_function}" ' +
        f'from "{register_function_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=register_function_path,
        resource='register_rpc',
        alias=register_function
    )
    task.print_out(f'Add "{register_function}" call to the code')
    code = append_code_to_function(
        code=code,
        function_name='register_rpc',
        new_code=f'{register_function}(logger, rpc_server, rpc_caller, publisher)' # noqa
    )
    task.print_out(f'Write modified code to: {module_rpc_file_path}')
    await write_text_file_async(module_rpc_file_path, code)


@typechecked
async def register_permission(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str
):
    module_register_permission_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', 'auth',
        'register_permission.py'
    )
    task.print_out(f'Read code from: {module_register_permission_file_path}')
    code = await read_text_file_async(module_register_permission_file_path)
    code = append_code_to_function(
        code=code,
        function_name='register_permission',
        new_code='\n'.join([
            'await ensure_entity_permission(',
            f"    module_name='{snake_module_name}', entity_name='{snake_entity_name}'", # noqa
            ')'
        ])
    )
    task.print_out(
        f'Add "ensure_entity_permission" call for {snake_entity_name} ' +
        'to the code'
    )
    task.print_out(
        f'Write modified code to: {module_register_permission_file_path}'
    )
    await write_text_file_async(module_register_permission_file_path, code)



from typing import Any, Mapping
from .group import project_group
from ..task.decorator import python_task
from ..task_input.bool_input import BoolInput
from ..runner import runner
from ..helper.env_map.fetch import fetch_env_map_from_group

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='get-default-env',
    description='Get default values for project environments',
    inputs=[
        BoolInput(
            name='export',
            shortcut='e',
            description='Whether add export statement or not',
            default=True
        )
    ],
    group=project_group,
    runner=runner
)
async def get_default_env(*args: Any, **kwargs: Any) -> str:
    env_map: Mapping[str, str] = {}
    env_map = fetch_env_map_from_group(env_map, project_group)
    env_keys = list(env_map.keys())
    env_keys.sort()
    should_export = kwargs.get('export', True)
    export_prefix = 'export ' if should_export else ''
    return '\n'.join([
        f'{export_prefix}{key}={env_map[key]}' for key in env_keys
    ])



from .devtool_install import (
    install_gvm, install_pyenv, install_nvm, install_sdkman, install_pulumi,
    install_aws, install_gcloud, install_tmux, install_zsh, install_kubectl,
    install_helm, install_docker, install_terraform, install_helix
)

assert install_gvm
assert install_pyenv
assert install_nvm
assert install_sdkman
assert install_pulumi
assert install_aws
assert install_gcloud
assert install_tmux
assert install_zsh
assert install_kubectl
assert install_helm
assert install_docker
assert install_terraform
assert install_helix



from typing import Any
from ..common.task_input import (
    project_dir_input, package_name_input, package_description_input,
    package_homepage_input, package_bug_tracker_input,
    package_author_name_input, package_author_email_input
)
from ..common.helper import (
    validate_existing_project_dir, validate_inexisting_automation
)
from ..common.task_factory import create_register_module
from ...group import project_add_group
from ....task.decorator import python_task
from ....task.task import Task
from ....task.resource_maker import ResourceMaker
from ....runner import runner

import os

CURRENT_DIR = os.path.dirname(__file__)


###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=[
        project_dir_input,
        package_name_input
    ]
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_existing_project_dir(project_dir)
    package_name = kwargs.get('package_name')
    validate_inexisting_automation(project_dir, package_name)


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        package_name_input,
        package_description_input,
        package_homepage_input,
        package_bug_tracker_input,
        package_author_name_input,
        package_author_email_input
    ],
    upstreams=[validate],
    replacements={
        'zrbPackageName': '{{input.package_name}}',
        'zrbPackageDescription': '{{input.package_description}}',
        'zrbPackageHomepage': '{{input.package_homepage}}',
        'zrbPackageBugTracker': '{{input.package_bug_tracker}}',
        'zrbPackageAuthorName': '{{input.package_author_name}}',
        'zrbPackageAuthorEmail': '{{input.package_author_email}}'
    },
    template_path=os.path.join(CURRENT_DIR, 'template'),
    destination_path='{{ input.project_dir }}',
    excludes=[
        '*/__pycache__',
    ]
)

register_module = create_register_module(
    module_path='_automate.{{util.to_snake_case(input.package_name)}}.local',
    alias='{{util.to_snake_case(input.package_name)}}_local',
    inputs=[package_name_input],
    upstreams=[copy_resource]
)


@python_task(
    name='pip-package',
    group=project_add_group,
    upstreams=[register_module],
    runner=runner
)
async def add_python_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from zrb import (
    runner, Env,
    StrInput, ChoiceInput, IntInput, BoolInput, FloatInput, PasswordInput,
    Group, Task, CmdTask, HTTPChecker, python_task
)

# Simple Python task.
# Usage example: zrb concat --separator=' '
concat = Task(
    name='concat',  # Task name
    inputs=[StrInput(name='separator', description='Separator', default=' ')],
    run=lambda *args, **kwargs: kwargs.get('separator', ' ').join(args)
)
runner.register(concat)

# Simple Python task with multiple inputs.
register_trainer = Task(
    name='register-trainer',
    inputs=[
        StrInput(name='name', default=''),
        PasswordInput(name='password', default=''),
        IntInput(name='age', default=0),
        BoolInput(name='employed', default=False),
        FloatInput(name='salary', default=0.0),
        ChoiceInput(
            name='starter-pokemon',
            choices=['bulbasaur', 'charmender', 'squirtle']
        )
    ],
    run=lambda *args, **kwargs: kwargs
)
runner.register(register_trainer)


# Simple Python task with decorator
@python_task(
    name='fibo',
    inputs=[IntInput(name='n', default=5)],
    runner=runner
)
async def fibo(*args, **kwargs):
    n = int(args[0]) if len(args) > 0 else kwargs.get('n')
    if n <= 0:
        return None
    elif n == 1:
        return 0
    elif n == 2:
        return 1
    else:
        a, b = 0, 1
        for i in range(n - 1):
            a, b = b, a + b
        return a


# Simple CLI task.
# Usage example: zrb hello --name='world'
hello = CmdTask(
    name='hello',
    inputs=[StrInput(name='name', description='Name', default='world')],
    cmd='echo Hello {{input.name}}'
)
runner.register(hello)

# Command group: zrb make
make = Group(name='make', description='Make things')

# CLI task, part of `zrb make` group, depends on `hello`
# Usage example: zrb make coffee
make_coffee = CmdTask(
    name='coffee',
    group=make,
    upstreams=[hello],
    cmd='echo Coffee for you ☕'
)
runner.register(make_coffee)

# CLI task, part of `zrb make` group, depends on `hello`
# Usage example: zrb make beer
make_beer = CmdTask(
    name='beer',
    group=make,
    upstreams=[hello],
    cmd='echo Cheers 🍺'
)
runner.register(make_beer)

# Command group: zrb make gitignore
make_gitignore = Group(
    name='gitignore', description='Make gitignore', parent=make
)

# CLI task, part of `zrb make gitignore` group,
# making .gitignore for Python project
# Usage example: zrb make gitignore python
make_gitignore_python = CmdTask(
    name='python',
    group=make_gitignore,
    cmd=[
        'echo "node_modules/" >> .gitignore'
        'echo ".npm" >> .gitignore'
        'echo "npm-debug.log" >> .gitignore'
    ]
)
runner.register(make_gitignore_python)

# CLI task, part of `zrb make gitignore` group,
# making .gitignore for Node.js project
# Usage example: zrb make gitignore node
make_gitignore_nodejs = CmdTask(
    name='node',
    group=make_gitignore,
    cmd=[
        'echo "__pycache__/" >> .gitignore'
        'echo "venv" >> .gitignore'
    ]
)
runner.register(make_gitignore_nodejs)

# Long running CLI task
# Usage example: zrb start-server dir='.'
start_server = CmdTask(
    name='start-server',
    upstreams=[make_coffee, make_beer],
    inputs=[StrInput(name='dir', description='Directory', default='.')],
    envs=[Env(name='PORT', os_name='WEB_PORT', default='3000')],
    cmd='python -m http.server $PORT --directory {{input.dir}}',
    checkers=[HTTPChecker(port='{{env.PORT}}')]
)
runner.register(start_server)

# CLI task, depends on `start-server`, throw error
# Usage example: zrb test-error
test_error = CmdTask(
    name='test-error',
    upstreams=[start_server],
    cmd='sleep 3 && exit 1',
    retry=0
)
runner.register(test_error)



from typing import List
from .helper.advertisement import Advertisement

# flake8: noqa E501

advertisements: List[Advertisement] = [
    Advertisement(
        content='\n'.join([
            'Support zrb growth and development!',
            '☕ Donate at: https://stalchmst.com/donation',
            '🐙 Submit issues/pull requests at: https://github.com/state-alchemists/zrb',
            '🐤 Follow us at: https://twitter.com/zarubastalchmst'
        ]),
        time_pattern='.*'
    ),
]



from . import common
from . import project_task
from .project.create import create_project
from .cmd_task.add import add_cmd_task
from .docker_compose_task.add import add_docker_compose_task
from .python_task.add import add_python_task
from .simple_python_app import add as add_simple_python_app
from .fastapp import add as add_fastapp
from .fastapp_module import add as add_fastapp_module
from .fastapp_crud import add as add_fastapp_crud
from .fastapp_field import add as add_fastapp_field
from .pip_package import add as add_pip_package
from .app_generator import add as add_app_generator

assert common
assert project_task
assert create_project
assert add_cmd_task
assert add_docker_compose_task
assert add_python_task
assert add_simple_python_app
assert add_fastapp
assert add_fastapp_module
assert add_fastapp_crud
assert add_fastapp_field
assert add_pip_package
assert add_app_generator



from typing import Any, List, Mapping, Optional, Union
from typeguard import typechecked
from .any_input import AnyInput
from .constant import RESERVED_INPUT_NAMES
from ..config.config import show_prompt


@typechecked
class BaseInput(AnyInput):

    def __init__(
        self,
        name: str,
        shortcut: Optional[str] = None,
        default: Optional[Any] = None,
        description: Optional[str] = None,
        show_default: Union[bool, str, None] = None,
        prompt: Union[bool, str] = True,
        confirmation_prompt: Union[bool, str] = False,
        prompt_required: bool = True,
        hide_input: bool = False,
        is_flag: Optional[bool] = None,
        flag_value: Optional[Any] = None,
        multiple: bool = False,
        count: bool = False,
        allow_from_autoenv: bool = True,
        type: Optional[Any] = None,
        hidden: bool = False,
        show_choices: bool = True,
        show_envvar: bool = False,
        nargs: int = 1,
    ):
        if name in RESERVED_INPUT_NAMES:
            raise ValueError(f'Forbidden input name: {name}')
        self._name = name
        self._shortcut = shortcut
        self._prompt = prompt
        self._default = default
        self._help = description if description is not None else name
        self._type = type
        self._show_default = show_default
        self._confirmation_prompt = confirmation_prompt
        self._prompt_required = prompt_required
        self._hide_input = hide_input
        self._is_flag = is_flag
        self._flag_value = flag_value
        self._multiple = multiple
        self._count = count
        self._allow_from_autoenv = allow_from_autoenv
        self._hidden = hidden
        self._show_choices = show_choices
        self._show_envvar = show_envvar
        self._nargs = nargs

    def get_name(self) -> str:
        return self._name

    def get_default(self) -> Any:
        return self._default

    def get_param_decl(self) -> List[str]:
        param_decl: List[str] = [f'--{self._name}']
        if self._shortcut is not None:
            param_decl.append(f'-{self._shortcut}')
        return param_decl

    def get_options(self) -> Mapping[str, Any]:
        options: Mapping[str, Any] = {
            'default': self._default,
            'help': self._help,
            'type': self._type,
            'show_default': self._show_default,
            'confirmation_prompt': self._confirmation_prompt,
            'prompt_required': self._prompt_required,
            'hide_input': self._hide_input,
            'is_flag': self._is_flag,
            'flag_value': self._flag_value,
            'multiple': self._multiple,
            'count': self._count,
            'allow_from_autoenv': self._allow_from_autoenv,
            'hidden': self._hidden,
            'show_choices': self._show_choices,
            'show_envvar': self._show_envvar,
            'nargs': self._nargs
        }
        if show_prompt:
            options['prompt'] = self._prompt
        return options



from typing import Any
from ..common.task_input import project_dir_input, task_name_input
from ..common.helper import (
    validate_existing_project_dir, validate_inexisting_automation
)
from ..common.task_factory import create_register_module
from ...group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task.resource_maker import ResourceMaker
from ....runner import runner

import os

CURRENT_DIR = os.path.dirname(__file__)

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=[
        project_dir_input,
        task_name_input,
    ]
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_existing_project_dir(project_dir)
    task_name = kwargs.get('task_name')
    validate_inexisting_automation(project_dir, task_name)


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        task_name_input,
    ],
    upstreams=[validate],
    replacements={
        'zrbTaskName': '{{input.task_name}}',
    },
    template_path=os.path.join(CURRENT_DIR, 'template'),
    destination_path='{{ input.project_dir }}',
    excludes=[
        '*/__pycache__',
    ]
)

register_module = create_register_module(
    module_path='_automate.{{util.to_snake_case(input.task_name)}}',
    inputs=[task_name_input],
    upstreams=[copy_resource]
)


@python_task(
    name='cmd-task',
    group=project_add_group,
    upstreams=[register_module],
    runner=runner
)
async def add_cmd_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from config import (
    app_enable_snake_zrb_module_name_module
)
from component.log import logger
from component.db_connection import engine
from helper.migration import migrate
from module.snake_zrb_module_name.component import Base


async def migrate_snake_zrb_module_name():
    if not app_enable_snake_zrb_module_name_module:
        logger.info('🥪 Skip DB migration for "snake_zrb_module_name"')
        return
    logger.info('🥪 Perform DB migration for "snake_zrb_module_name"')
    await migrate(engine=engine, Base=Base)



from typing import Any, List
from .group import env_group
from ..helper.accessories.color import colored
from ..task.decorator import python_task
from ..task.task import Task
from ..runner import runner

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='get',
    group=env_group,
    description='Get environment values',
    runner=runner
)
async def get(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    env_map = task.get_env_map()
    names = list(env_map.keys())
    names.sort()
    colored_equal = colored('=', color='grey', attrs=['dark'])
    env_lines: List[str] = []
    for name in names:
        value = env_map[name]
        colored_name = colored(name, color='green', attrs=['bold'])
        colored_value = colored(value, attrs=['bold'])
        env_lines.append(f'{colored_name}{colored_equal}{colored_value}')
    line_separator = '\n    '
    task.print_out(line_separator + line_separator.join(env_lines))
    return env_map



from typing import Any
from .group import git_group
from ..task.decorator import python_task
from ..task.task import Task
from ..task_input.str_input import StrInput
from ..runner import runner
from ..helper.git.detect_changes import get_modified_files
from ..helper.accessories.color import colored


###############################################################################
# Input Definitions
###############################################################################
commit_input = StrInput(
    name='commit',
    shortcut='c',
    description='commit hash/tag',
    default='HEAD'
)

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='get-changes',
    group=git_group,
    description='Get modified files',
    inputs=[commit_input],
    runner=runner
)
async def show(*args: Any, **kwargs: Any):
    commit = kwargs.get('commit', 'HEAD')
    task: Task = kwargs['_task']
    modified_files = get_modified_files(commit)
    modified_file_keys = list(modified_files.keys())
    modified_file_keys.sort()
    for modified_file, state in modified_files.items():
        if state.minus and state.plus:
            task.print_out(colored(f'+- {modified_file}', color='yellow'))
            continue
        if state.minus:
            task.print_out(colored(f'-- {modified_file}', color='red'))
            continue
        if state.plus:
            task.print_out(colored(f'++ {modified_file}', color='green'))
            continue
    return '\n'.join(modified_file_keys)



from typeguard import typechecked
from typing import Mapping
import subprocess


@typechecked
class ModificationState():
    def __init__(self):
        self.plus: bool = False
        self.minus: bool = False


@typechecked
def get_modified_files(commit: str) -> Mapping[str, ModificationState]:
    exit_status, output = subprocess.getstatusoutput(f'git show {commit}')
    if exit_status != 0:
        raise Exception(output)
    lines = output.split('\n')
    modified_files: Mapping[str, ModificationState] = {}
    for line in lines:
        if (
            not line.startswith('---') and
            not line.startswith('+++') and
            line[4:6] != 'a/' and
            line[4:6] != 'b/'
        ):
            continue
        # line should contains something like `--- a/some-file.txt`
        file = line[6:]
        if file not in modified_files:
            modified_files[file] = ModificationState()
        modification_state = modified_files[file]
        if line.startswith('---'):
            modification_state.minus = True
        if line.startswith('+++'):
            modification_state.plus = True
    return modified_files


from typing import Any
from zrb import (
    BoolInput, ChoiceInput, StrInput, Env, HTTPChecker, PortChecker
)
import jsons
import os

###############################################################################
# Constants
###############################################################################

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, '..', '..'))
RESOURCE_DIR = os.path.join(PROJECT_DIR, 'src', 'kebab-zrb-app-name')
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, 'deployment')
DEPLOYMENT_TEMPLATE_ENV_FILE_NAME = os.path.join(
    DEPLOYMENT_DIR, 'template.env'
)
APP_DIR = os.path.join(RESOURCE_DIR, 'src')
APP_FRONTEND_DIR = os.path.join(APP_DIR, 'frontend')
APP_FRONTEND_BUILD_DIR = os.path.join(APP_FRONTEND_DIR, 'build')
APP_TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, 'template.env')
MODULE_CONFIG_PATH = os.path.join(CURRENT_DIR, 'config', 'modules.json')
with open(MODULE_CONFIG_PATH) as file:
    MODULE_JSON_STR = file.read()
MODULES = jsons.loads(MODULE_JSON_STR)

###############################################################################
# Functions
###############################################################################


def skip_local_microservices_execution(*args: Any, **kwargs: Any) -> bool:
    if not kwargs.get('local_snake_zrb_app_name', True):
        return True
    run_mode = kwargs.get('snake_zrb_app_name_run_mode', 'monolith')
    return run_mode != 'microservices'


###############################################################################
# Checker Task Definitions
###############################################################################

rabbitmq_management_checker = HTTPChecker(
    name='check-rabbitmq-management',
    port='{{env.get("RABBITMQ_MANAGEMENT_HOST_PORT", "15672")}}',
    is_https='{{input.snake_zrb_app_name_https}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "rabbitmq"}}'
)

rabbitmq_checker = PortChecker(
    name='check-rabbitmq',
    port='{{env.get("RABBITMQ_HOST_PORT", "5672")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "rabbitmq"}}'
)

redpanda_console_checker = HTTPChecker(
    name='check-redpanda-console',
    method='GET',
    port='{{env.get("REDPANDA_CONSOLE_HOST_PORT", "9000")}}',
    is_https='{{input.snake_zrb_app_name_https}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

kafka_plaintext_checker = PortChecker(
    name='check-kafka-plaintext',
    port='{{env.get("KAFKA_INTERNAL_HOST_PORT", "29092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

kafka_outside_checker = PortChecker(
    name='check-kafka-outside',
    port='{{env.get("KAFKA_EXTERNAL_HOST_PORT", "9092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

pandaproxy_plaintext_checker = PortChecker(
    name='check-pandaproxy-plaintext',
    port='{{env.get("PANDAPROXY_INTERNAL_HOST_PORT", "29092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

pandaproxy_outside_checker = PortChecker(
    name='check-pandaproxy-outside',
    port='{{env.get("PANDAPROXY_EXTERNAL_HOST_PORT", "9092")}}',
    skip_execution='{{env.get("APP_BROKER_TYPE", "rabbitmq") != "kafka"}}'
)

app_container_checker = HTTPChecker(
    name='check-kebab-zrb-app-name-container',
    host='{{input.snake_zrb_app_name_host}}',
    url='/readiness',
    port='{{env.get("HOST_PORT", "zrbAppHttpPort")}}',
    is_https='{{input.snake_zrb_app_name_https}}'
)

app_local_checker = HTTPChecker(
    name='check-kebab-zrb-app-name',
    host='{{input.snake_zrb_app_name_host}}',
    url='/readiness',
    port='{{env.APP_PORT}}',
    is_https='{{input.snake_zrb_app_name_https}}',
    skip_execution=skip_local_microservices_execution
)

###############################################################################
# Input Definitions
###############################################################################

enable_monitoring_input = BoolInput(
    name='enable-kebab-zrb-app-name-monitoring',
    description='Enable "kebab-zrb-app-name" monitoring',
    prompt='Enable "kebab-zrb-app-name" monitoring?',
    default=False
)

local_input = BoolInput(
    name='local-kebab-zrb-app-name',
    description='Use "kebab-zrb-app-name" on local machine',
    prompt='Use "kebab-zrb-app-name" on local machine?',
    default=True
)

run_mode_input = ChoiceInput(
    name='kebab-zrb-app-name-run-mode',
    description='"kebab-zrb-app-name" run mode (monolith/microservices)',
    prompt='Run "kebab-zrb-app-name" as a monolith or microservices?',
    choices=['monolith', 'microservices'],
    default='monolith'
)

https_input = BoolInput(
    name='kebab-zrb-app-name-https',
    description='Whether "kebab-zrb-app-name" run on HTTPS',
    prompt='Is "kebab-zrb-app-name" run on HTTPS?',
    default=False
)

host_input = StrInput(
    name='kebab-zrb-app-name-host',
    description='Hostname of "kebab-zrb-app-name"',
    prompt='Hostname of "kebab-zrb-app-name"',
    default='localhost'
)

###############################################################################
# Env fDefinitions
###############################################################################

app_enable_otel_env = Env(
    name='APP_ENABLE_OTEL',
    default='{{ "1" if input.enable_snake_zrb_app_name_monitoring else "0" }}'
)



import os

message = os.getenv('MY_MESSAGE', 'Hello, world!')
post_message_1 = os.getenv('POST_MESSAGE_1', '')
post_message_2 = os.getenv('POST_MESSAGE_2', '')

print(message)
if post_message_1 != '':
    print(post_message_1)
if post_message_2 != '':
    print(post_message_2)



from typing import Any
from .group import base64_group
from ..task.decorator import python_task
from ..task_input.str_input import StrInput
from ..runner import runner

import base64

###############################################################################
# Input Definitions
###############################################################################

text_input = StrInput(
    name='text',
    shortcut='t',
    description='Text',
    default=''
)

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='encode',
    group=base64_group,
    inputs=[text_input],
    description='Encode base64 task',
    retry=0,
    runner=runner
)
async def encode(*args: str, **kwargs: Any):
    text: str = kwargs.get('text', '')
    encoded_text = base64.b64encode(text.encode())
    return encoded_text.decode()


@python_task(
    name='decode',
    group=base64_group,
    inputs=[text_input],
    description='Decode base64 task',
    retry=0,
    runner=runner
)
async def decode(*args: str, **kwargs: Any):
    text: str = kwargs.get('text', '')
    encoded_text = base64.b64decode(text.encode())
    return encoded_text.decode()



from typing import Any
from ..common.task_input import project_dir_input
from ..common.helper import (
    validate_existing_project_dir, validate_inexisting_automation
)
from ..common.task_factory import create_register_module
from ...group import project_add_group
from ....task.decorator import python_task
from ....task.task import Task
from ....task.resource_maker import ResourceMaker
from ....task_input.str_input import StrInput
from ....helper.accessories.name import get_random_name
from ....runner import runner

import os

CURRENT_DIR = os.path.dirname(__file__)


template_name_input = StrInput(
    name='template-name',
    shortcut='t',
    description='Template name',
    prompt='Template name',
    default=get_random_name()
)

template_base_image_input = StrInput(
    name='template-base-image',
    shortcut='i',
    prompt='Base image',
    default='python:3.10-slim'
)


###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=[
        project_dir_input,
        template_name_input
    ]
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_existing_project_dir(project_dir)
    template_name = kwargs.get('template_name')
    validate_inexisting_automation(project_dir, template_name)


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        template_name_input,
        template_base_image_input,
    ],
    upstreams=[validate],
    replacements={
        'zrbMetaTemplateName': '{{input.template_name}}',
        'zrbMetaTemplateBaseImage': '{{input.template_base_image}}',
    },
    template_path=os.path.join(CURRENT_DIR, 'template'),
    destination_path='{{ input.project_dir }}',
    excludes=[
        '*/__pycache__',
    ]
)

register_module = create_register_module(
    module_path='_automate.{{util.to_snake_case(input.template_name)}}.add',
    alias='{{util.to_snake_case(input.template_name)}}_add',
    inputs=[template_name_input],
    upstreams=[copy_resource]
)


@python_task(
    name='app-generator',
    group=project_add_group,
    upstreams=[register_module],
    runner=runner
)
async def add_python_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from alembic.migration import MigrationContext
from alembic.autogenerate import produce_migrations
from alembic.operations.ops import ModifyTableOps
from alembic.operations import Operations
from sqlalchemy import Engine
from sqlalchemy.orm import DeclarativeBase


async def migrate(engine: Engine, Base: DeclarativeBase):
    '''
    Generate migration and run it
    See: https://alembic.sqlalchemy.org/en/latest/cookbook.html#run-alembic-operation-objects-directly-as-in-from-autogenerate  # noqa
    '''
    connection = engine.connect()
    migration_context = MigrationContext.configure(connection)
    migration_script = produce_migrations(
        context=migration_context,
        metadata=Base.metadata
    )
    operations = Operations(migration_context)
    use_batch = engine.name == "sqlite"

    stack = [migration_script.upgrade_ops]
    while stack:
        elem = stack.pop(0)
        if (
            hasattr(elem, 'table_name') and
            elem.table_name not in Base.metadata.tables
        ):
            # We want the migration leave all unrelated tables as is
            continue

        if use_batch and isinstance(elem, ModifyTableOps):
            with operations.batch_alter_table(
                elem.table_name, schema=elem.schema
            ) as batch_ops:
                for table_elem in elem.ops:
                    # work around Alembic issue #753 (fixed in 1.5.0)
                    if hasattr(table_elem, "column"):
                        table_elem.column = table_elem.column.copy()
                    batch_ops.invoke(table_elem)
        elif hasattr(elem, "ops"):
            stack.extend(elem.ops)
        else:
            # work around Alembic issue #753 (fixed in 1.5.0)
            if hasattr(elem, "column"):
                elem.column = elem.column.copy()
            operations.invoke(elem)
    connection.commit()
    connection.close()



from typing import List, Optional, TypeVar
from typeguard import typechecked
from ..task.any_task import AnyTask
from ..helper.string.conversion import to_cmd_name

TGroup = TypeVar('TGroup', bound='Group')


@typechecked
class Group():
    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        parent: Optional[TGroup] = None
    ):
        self._name = name
        self._description = description
        self._parent = parent
        if parent is not None:
            parent._children.append(self)
        self._children: List[TGroup] = []
        self._tasks: List[AnyTask] = []

    def get_cmd_name(self) -> str:
        return to_cmd_name(self._name)

    def get_complete_name(self) -> str:
        cmd_name = self.get_cmd_name()
        if self._parent is None:
            return cmd_name
        parent_cmd_name = self._parent.get_complete_name()
        return f'{parent_cmd_name} {cmd_name}'

    def get_id(self) -> str:
        group_id = self.get_cmd_name()
        if self._parent is None:
            return group_id
        parent_group_id = self._parent.get_id()
        return f'{parent_group_id} {group_id}'

    def add_task(self, task: AnyTask):
        self._tasks.append(task)

    def get_tasks(self) -> List[AnyTask]:
        return self._tasks

    def get_children(self) -> List[TGroup]:
        return self._children



from typeguard import typechecked
from ....task.task import Task
from ....helper.codemod.add_property_to_class import add_property_to_class
from ....helper.codemod.add_key_value_to_dict import add_key_value_to_dict
from ....helper.file.text import read_text_file_async, write_text_file_async
import os
import re


@typechecked
async def add_column_to_insert_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'new', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add default value to insert page')
    default_value_regex = r"(.*)(// DON'T DELETE: set field default value here)"  # noqa
    default_value_subst = '\\n'.join([
        f"\\1row.{snake_column_name} = '';",
        '\\1\\2'
    ])
    html_content = re.sub(
        default_value_regex, default_value_subst, html_content, 0, re.MULTILINE
    )
    task.print_out('Add field to insert page')
    field_regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    field_subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <input type="text" class="input w-full" id="{kebab_column_name}" placeholder="{capitalized_human_readable_column_name}" bind:value=' + '{row.' + snake_column_name + '} />', # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        field_regex, field_subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


@typechecked
async def add_column_to_update_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'update', '[id]', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add field to update page')
    regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <input type="text" class="input w-full" id="{kebab_column_name}" placeholder="{capitalized_human_readable_column_name}" bind:value=' + '{row.' + snake_column_name + '} />', # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        regex, subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


@typechecked
async def add_column_to_delete_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'delete', '[id]', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add field to delete page')
    regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <span id="{kebab_column_name}">' + '{row.' + snake_column_name + '}</span>', # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        regex, subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


@typechecked
async def add_column_to_detail_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'detail', '[id]', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add field to detail page')
    regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <span id="{kebab_column_name}">' + '{row.' + snake_column_name + '}</span>',  # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        regex, subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


@typechecked
async def add_column_to_list_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    # process header
    task.print_out('Add column header to table')
    header_regex = r"(.*)(<!-- DON'T DELETE: insert new column header here-->)"
    header_subst = '\\n'.join([
        f'\\1<th>{capitalized_human_readable_column_name}</th>',
        '\\1\\2'
    ])
    html_content = re.sub(
        header_regex, header_subst, html_content, 0, re.MULTILINE
    )
    # process column
    task.print_out('Add column to table')
    column_regex = r"(.*)(<!-- DON'T DELETE: insert new column here-->)"
    column_subst = '\\n'.join([
        '\\1<td>{row.' + snake_column_name + '}</td>',
        '\\1\\2'
    ])
    html_content = re.sub(
        column_regex, column_subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


@typechecked
async def add_column_to_test(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    test_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'test', snake_module_name,
        f'test_{snake_entity_name}.py'
    )
    task.print_out(f'Read code from: {test_file_path}')
    code = await read_text_file_async(test_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the test'
    )
    dict_names = [
        'inserted_success_data', 'to_be_updated_success_data',
        'updated_success_data', 'to_be_deleted_success_data'
    ]
    for dict_name in dict_names:
        code = add_key_value_to_dict(
            code=code,
            dict_name=dict_name,
            key=f"'{snake_column_name}'",
            value="''"
        )
    task.print_out(f'Write modified code to: {test_file_path}')
    await write_text_file_async(test_file_path, code)


@typechecked
async def add_column_to_schema(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    pascal_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    schema_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'schema', f'{snake_entity_name}.py'
    )
    task.print_out(f'Read code from: {schema_file_path}')
    code = await read_text_file_async(schema_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the schema'
    )
    code = add_property_to_class(
        code=code,
        class_name=f'{pascal_entity_name}Data',
        property_name=snake_column_name,
        property_type='str'
    )
    task.print_out(f'Write modified code to: {schema_file_path}')
    await write_text_file_async(schema_file_path, code)


@typechecked
async def add_column_to_repo(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    pascal_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    repo_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'entity', snake_entity_name, 'repo.py'
    )
    task.print_out(f'Read code from: {repo_file_path}')
    code = await read_text_file_async(repo_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the repo'
    )
    code = add_property_to_class(
        code=code,
        class_name=f'DBEntity{pascal_entity_name}',
        property_name=snake_column_name,
        property_type='Column',
        property_value='Column(String)'
    )
    task.print_out(f'Write modified code to: {repo_file_path}')
    await write_text_file_async(repo_file_path, code)



from zrb.task.docker_compose_task import DockerComposeTask, ServiceConfig, Env, EnvFile

import pathlib
import os


def test_docker_compose_task_simple():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple',
        compose_file=os.path.join(resource_path, 'docker-compose-simple.yml'),
        compose_service_configs={
            'myapp_simple': ServiceConfig(
                envs=[Env(name='POST_MESSAGE_2', default='As below so above')],
                env_files=[EnvFile(os.path.join(resource_path, 'runtime.env'))]
            )
        },
        compose_flags=['--build']
    )
    function = docker_compose_task.to_function()
    result = function()
    assert 'Hello, Compose!' in result.output
    assert 'As above so below' in result.output
    assert 'As below so above' in result.output


def test_docker_compose_task_with_cwd_and_compose_file():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='with-cwd-and-compose-file',
        cwd=resource_path,
        compose_file='docker-compose-simple.yml',
        compose_service_configs={
            'myapp_simple': ServiceConfig(
                envs=[Env(name='POST_MESSAGE_2', default='As below so above')],
                env_files=[EnvFile(os.path.join(resource_path, 'runtime.env'))]
            )
        },
        compose_flags=['--build']
    )
    function = docker_compose_task.to_function()
    result = function()
    assert 'Hello, Compose!' in result.output
    assert 'As above so below' in result.output
    assert 'As below so above' in result.output


def test_docker_compose_task_with_cwd():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='with-cwd',
        cwd=resource_path,
        compose_service_configs={
            'myapp': ServiceConfig(
                envs=[Env(name='POST_MESSAGE_2', default='As below so above')],
                env_files=[EnvFile(os.path.join(resource_path, 'runtime.env'))]
            )
        },
        compose_flags=['--build']
    )
    function = docker_compose_task.to_function()
    result = function()
    assert 'Hello, Compose!' in result.output
    assert 'As above so below' in result.output
    assert 'As below so above' in result.output


def test_docker_compose_task_invalid_compose_file():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    is_error: bool = False
    try:
        DockerComposeTask(
            name='simple',
            cwd=resource_path,
            compose_file='non-existing.yml',
        )
    except Exception:
        is_error = True
    assert is_error


def test_docker_compose_task_simple_no_default_env():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple-no-default-env',
        compose_file=os.path.join(
            resource_path, 'docker-compose-simple-no-default-env.yml'
        ),
        compose_service_configs={
            'myapp_simple_no_default_env': ServiceConfig(
                envs=[Env(name='POST_MESSAGE_2', default='As below so above')],
                env_files=[EnvFile(os.path.join(resource_path, 'runtime.env'))]
            )
        },
        compose_flags=['--build']
    )
    function = docker_compose_task.to_function()
    # run with env set
    if 'ZRB_TEST_DC_TASK_SIMPLE_NO_DEFAULT_MESSAGE' in os.environ:
        del os.environ['ZRB_TEST_DC_TASK_SIMPLE_NO_DEFAULT_MESSAGE']
    os.environ['ZRB_TEST_DC_TASK_SIMPLE_NO_DEFAULT_MESSAGE'] = 'Good night'
    result_default = function()
    assert 'Good night' in result_default.output
    assert 'As above so below' in result_default.output
    assert 'As below so above' in result_default.output
    del os.environ['ZRB_TEST_DC_TASK_SIMPLE_NO_DEFAULT_MESSAGE']


def test_docker_compose_task_simple_map_env():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple-map-env',
        compose_file=os.path.join(
            resource_path, 'docker-compose-simple-map-env.yml'
        ),
        compose_service_configs={
            'myapp_simple_map_env': ServiceConfig(
                envs=[Env(name='POST_MESSAGE_2', default='As below so above')],
                env_files=[EnvFile(os.path.join(resource_path, 'runtime.env'))]
            )
        },
        compose_flags=['--build']
    )
    function = docker_compose_task.to_function()
    # run with no env
    if 'ZRB_TEST_DC_TASK_SIMPLE_MAP_MESSAGE' in os.environ:
        del os.environ['ZRB_TEST_DC_TASK_SIMPLE_MAP_MESSAGE']
    result_default = function()
    assert 'Good morning' in result_default.output
    # run with env set to empty string
    os.environ['ZRB_TEST_DC_TASK_SIMPLE_MAP_MESSAGE'] = ''
    result_default = function()
    assert 'Good morning' in result_default.output
    # run with env set
    os.environ['ZRB_TEST_DC_TASK_SIMPLE_MAP_MESSAGE'] = 'Good night'
    result_custom = function()
    assert 'Good night' in result_custom.output
    del os.environ['ZRB_TEST_DC_TASK_SIMPLE_MAP_MESSAGE']


def test_docker_compose_task_simple_list_env():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple-list-env',
        compose_file=os.path.join(
            resource_path, 'docker-compose-simple-list-env.yml'
        ),
        compose_service_configs={
            'myapp_simple_list_env': ServiceConfig(
                envs=[Env(name='POST_MESSAGE_2', default='As below so above')],
                env_files=[EnvFile(os.path.join(resource_path, 'runtime.env'))]
            )
        },
        compose_flags=['--build']
    )
    function = docker_compose_task.to_function()
    # run with no env
    if 'ZRB_TEST_DC_TASK_SIMPLE_LIST_MESSAGE' in os.environ:
        del os.environ['ZRB_TEST_DC_TASK_SIMPLE_LIST_MESSAGE']
    result_default = function()
    assert 'Good morning' in result_default.output
    # run with env set to empty string
    os.environ['ZRB_TEST_DC_TASK_SIMPLE_LIST_MESSAGE'] = ''
    result_default = function()
    assert 'Good morning' in result_default.output
    # run with env set
    os.environ['ZRB_TEST_DC_TASK_SIMPLE_LIST_MESSAGE'] = 'Good night'
    result_custom = function()
    assert 'Good night' in result_custom.output
    del os.environ['ZRB_TEST_DC_TASK_SIMPLE_LIST_MESSAGE']



from zrb.builtin.explain import (
    explain_dry, explain_kiss, explain_yagni, explain_solid,
    explain_zen_of_python
)


def test_explain_dry():
    function = explain_dry.to_function()
    result = function()
    assert result is None


def test_explain_kiss():
    function = explain_kiss.to_function()
    result = function()
    assert result is None


def test_explain_solid():
    function = explain_solid.to_function()
    result = function()
    assert result is None


def test_explain_yagni():
    function = explain_yagni.to_function()
    result = function()
    assert result is None


def test_explain_zen_of_python():
    function = explain_zen_of_python.to_function()
    result = function()
    assert result is None



from typing import Any, List, Mapping
from abc import ABC, abstractmethod


class AnyInput(ABC):
    '''
    Task Input specification
    '''
    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def get_default(self) -> Any:
        pass

    @abstractmethod
    def get_param_decl(self) -> List[str]:
        pass

    @abstractmethod
    def get_options(self) -> Mapping[str, Any]:
        pass



from . import group
from . import base64
from . import env
from . import eval
from . import git
from . import md5
from . import explain
from . import project
from . import ubuntu
from . import update
from . import version
from . import devtool
from . import generator

assert group
assert base64
assert env
assert eval
assert git
assert md5
assert explain
assert project
assert ubuntu
assert update
assert version
assert devtool
assert generator



from zrb.builtin.generator.project.create import create_project
from zrb.builtin.generator.docker_compose_task.add import (
    add_docker_compose_task
)
import os
import pathlib
import shutil


def test_add_docker_compose_task():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    project_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(project_path):
        shutil.rmtree(project_path)

    create_project_loop = create_project.to_function()
    create_project_loop(project_dir=project_path)

    automate_path = os.path.join(project_path, '_automate')
    src_path = os.path.join(project_path, 'src')

    # first attempt should success
    first_attempt_loop = add_docker_compose_task.to_function()
    first_attempt_loop(
        project_dir=project_path, task_name='composeTask'
    )

    # compose_task.py file should exists
    assert os.path.isfile(
        os.path.join(automate_path, 'compose_task.py')
    )
    assert os.path.isfile(
        os.path.join(src_path, 'compose-task', 'docker-compose.yml')
    )

    # compose_task should be imported
    with open(os.path.join(project_path, 'zrb_init.py')) as f:
        content = f.read()
    assert 'assert compose_task' in content
    assert 'import _automate.compose_task as compose_task' in content

    # second attempt should fail
    is_error = False
    try:
        second_attempt_loop = add_docker_compose_task.to_function()
        second_attempt_loop(
            project_dir=project_path, task_name='composeTask'
        )
    except Exception:
        is_error = True
    assert is_error
